#!/usr/bin/env python
# coding:utf-8
"""
  This file is part of SoftKnit21 framework

License
=======

 Copyright (C) <2019>  <Odile Troulet-Lamberttex@orange.fr>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or  any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.



Module purpose
==============

LC Message is an abstract class ; LCMessageReceived and LCMessagetoSend inherit from LCMessage
Implements
==========

 - B{LCMessage}

Documentation
=============

    LCMessages and Payload classes

Usage
=====
exceptions raised only for errors from programming environment
errors which might come from transmission on serial port are not raised but logged at critical level

@author: Odile Troulet-Lambert
@copyright: Odile Troulet-Lambert
@license: AGPL
"""


from abc import ABC, abstractmethod
import sys


class Softknit21Exception (Exception):
    """parent class of all exceptions of the programm"""
    pass


class LCMessageError (Softknit21Exception):
    """ error in a LCmessage"""

    def __init__(self, errormessage, LCmessage):
        """ any error found in received messages or sent messages is captured here
        to get more flexibility in handling these errors"""
        pass


class Payload (list):
    """ Payload list containing only integer representing the bytes of the payload
    subclass of list to prevent from adding incorrect type values
    """

    def __init__(self):
        super().__init__()
        self.byteorder = sys.byteorder

    def pvalue(self, oneval):
        """returns the value of one byte of the payload"""
        if isinstance(oneval, bytes):
            return int.from_bytes(oneval, byteorder=self.byteorder)
        elif isinstance(oneval, int):
            return oneval
        else:
            LCMessageError(f'{oneval} is not an integer nor a byte', self)

    def add(self, value):
        if isinstance(value, list):
            for e in value:
                super().append(self.pvalue(e))
        else:
            super().append(self.pvalue(value))

    def append(self, value):

        LCMessageError(f'cannot append to the payload, use addpayload', self)

    def insert(self, value):
        LCMessageError(f'cannot append to the payload, us addpayload', self)

    def extend(self, value):
        LCMessageError(f'cannot append to the payload, us addpayload', self)

    def __str__(self):
        s = []
        for e in self:
            s.append(str(hex(e)))
        return str(s)


class LCMessage (ABC):
    """ contains a LCMessage

    @att mtype : type of the message
    @type : ascii character

    @att lenpayload : length of the payload of the message
    @type : int (-1 for a single byte message)

    @att payload : payload of the message
    @type : list of integers

    @att checksum : checked checksum
    @type : True or False

    """
    @abstractmethod
    def __init__(self):
        """ constructor"""
        pass

    @property
    def checksum(self):
        if self.mtype.islower():
            return self._checksum
        else:
            return None

    def set_checksum(self, value=None):
        # type has lost its MSB set to one
        # do not care because checksum is modulo 128

        if self.lenpayload <= 0 or self.lenpayload == None:
            # a singlebyte message has no checksum
            LCMessageError(f'type : {self.mtype} checksum proposed {value } : no checksum for single byte message', self)

        check = ord(self.mtype) + self.lenpayload
        for b in self.payload:
            if b is not None:
                check = check + b
        # if no checksum message to send, calculate the checksum and store it as an attribute
        if value is None and not self.received:
            self._checksum = (check % 128)
            return
        if isinstance(value, bytes):
            value = int.from_bytes(value, self.payload.byteorder)
        self._checksum = (check % 128 == value)
        return

    @property
    def payload(self):
        return self._payload

    @payload.setter
    def payload(self, value):
        # used only to set a new list as payload
        # in the form : self.payload = value
        if (isinstance(value, list)
                and len(value) == self.lenpayload
                and self.mtype.islower()
                ):
            self._payload = Payload()
            self._payload.add(value)
        # for a single byte only empty list for payload
        elif (self.mtype.isupper() and
              self.lenpayload == -1 and
              isinstance(value, list) and
              len(value) == 0
              ):

            self._payload = Payload()
        else:
            LCMessageError(f'payload must be a list', self)

    def isValid(self):
        """ checks the validity of the message
        can only apply on complete messages"""
        if not isinstance(self.mtype, str):
               return False
        elif self.mtype.isupper():
            return (self.lenpayload == -1
                    and len(self.payload) == 0
                    and self.checksum == None
                    )
        elif self.mtype.islower() and self.received:
            return (self.lenpayload >= 1 and
                    len(self.payload) == self.lenpayload and
                    isinstance(self.checksum, bool)
                    )
        elif self.mtype.islower() and not self.received:
            return (self.lenpayload >= 1 and
                    len(self.payload) == self.lenpayload and
                    isinstance(self.checksum, int)
                    )
        else:
             return False

    def _extra(self, dic, value):
        """ applies a modification to value
        @param dic: describes the modification to apply
        @type : dic

        @param value : the value to be modified"""
        import re
        if not isinstance(dic, dict):
            LCMessageError(f'extra is not a dict : {dic}', self)
            return value
        receive = dic.get('receive')
        if receive is not None:
            dic = dic['receive']
        li = dic.get('checklist')
        if li is not None:
            if value not in li:
                LCMessageError(f'value {value} is not in dic : {li}', self)
            return value
        factor = dic.get('factor')
        if factor is not None:
            if self.received:
                return value / factor
            if not self.received:
                return int(value * factor)
        offset = dic.get('offset')
        if offset is not None:
            if self.received:
                return value - offset
            if not self.received:
                return value + offset
        islist = dic.get('islist')
        if islist is not None and isinstance(islist, list):
            if self.received and isinstance(value, list):
                return dict(list(zip(islist, value)))

            if not self.received:
                # in self.datadict get the list of values in the right order
                orderedvalues = [self.datadict.get(key) for key in islist]
                if len(orderedvalues) != len(islist) or (
                    any(e for e in orderedvalues if (e != 0 and e != 1))
                ):
                    LCMessageError(f'error in list of bits values {orderedvalues}', self)
                return orderedvalues
        valid = dic.get('checkconf')
        if valid is not None and callable(valid):
            res = valid.__call__(value)
            if res != True:
                LCMessageConfError(f'error in lock controler configuration {res}', self)
                return value
            else:
                return value
        regexp = dic.get('checkregexp')
        if regexp is not None:
            res = re.search(regexp, value)
            if res is not None:
                s = res.span()
                if not(s[0] == 0 and s[1] == len(value)):
                    LCMessageError(f'value is not a O1 string {value}', self)
            return value
        sliceO = dic.get('sliceO')
        if sliceO is not None:
            value = value[sliceO]
            return value

        # here invalid dic entry
        LCMessageError(f'extra {dic} could not be interpreted', self)
        return value

    def __str__(self):
        """ returns a string representation of LCmessage with hex in payload"""

        if self.mtype.isupper() and self.payload == -1:
            return f"type : {self.mtype}"
        elif self.isValid():
            s = self.payload.__str__()
            return (f"type : {self.mtype} len : {self.lenpayload} "
                     f"payload : {s} "
                     f"checksum :{self._checksum}"
                    )
        else:
            s = self.payload.__str__()
            return (f"invalid message type : {self.mtype} len : {self.lenpayload} "
                     f"payload : {s} "
                     f" checksum : {self._checksum}"
                    )

    def strucLogMessage(self):
        s = StrucLogMessage(mtype=self.mtype,
                            lenpayload=self.lenpayload,
                            payload=self.payload,
                            checksum=self._checksum,
                            received=self.received,
                            ** self.datadict)

        return s

    def strucLogEvent(self):
        s = StrucLogEvent(name=self.name,
                          checksum=self._checksum,
                          received=self.received,
                          ** self.datadict)
        return s

    def toDict(self):

        return {'mtype': self.mtype,
                'lenpayload': self.lenpayload,
                'payload': self.payload,
                'checksum': self.checksum,
                'received': self.received,
                **self.datadict}

    def comp(self, dic):

        def _cmpitem(a, b):
            """ returns True if payload items a and b are equal """
            if isinstance(a, bytes) and isinstance(b, int):
                return int.from_bytes(a, self.payload.byteorder) == b
            if isinstance(b, bytes) and isinstance(a, int):
                return int.from_bytes(b, self.payload.byteorder) == a
            if isinstance(a, int) and isinstance(b, int):
                return a == b
            if a is None and b is None:
                return True
            LCMessageError(f'cannot compare {a} and {b}', self)

        def _payloadcmp(payload1, payload2):
            """ compares 2 payloads of LCMessage"""
            if (isinstance(payload1, list) and
                        isinstance(payload2, list) and
                        len(payload1) == len(payload2)
                    ) :
                return all(_cmpitem(e1, e2) for e1, e2 in zip(payload1, payload2))

        dic2 = self.toDict()
        exceptpayload = all(dic2.get(key) == value for key, value in dic.items() if key != 'payload' and key != 'checksum')
        if dic.get('checksum') != None:
            checksum = _cmpitem(dic2.get('checksum'), dic.get('checksum'))
        else:
            checksum = True
        payload = _payloadcmp(dic2.get('payload'), dic.get('payload'))
        return exceptpayload and payload and checksum


if __name__ == '__main__':

    import unittest
    import time
    from unittest.mock import patch, Mock
    # no constructor for abstract messages tested throught LC messagereceived

    unittest.main()


