#!/usr/bin/env python
# coding:utf-8
"""
  This file is part of SoftKnit21 framework

License
=======

 Copyright (C) <2019>  <Odile Troulet-Lamberttex@orange.fr>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU Affero General Public License as
    published by the Free Software Foundation, either version 3 of the
    License, or  any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Affero General Public License for more details.

    You should have received a copy of the GNU Affero General Public License
    along with this program.  If not, see <https://www.gnu.org/licenses/>.



Module purpose
==============

LC messages received from the lock controler and their payload

Implements
==========

 - B{Lock Controler Manager}

Documentation
=============

    LC messages as received by the lock controler
Usage
=====


@author: Odile Troulet-Lambert
@copyright: Odile Troulet-Lambert
@license: AGPL
"""


from LC_message_abstract import LCMessage, Payload
from LC_message_abstract import LCMessageError


class LCMessageReceived (LCMessage):
    """ contains a received message
    @att mtype : type of the message
    @type : ascii character

    @att lenpayload : length of the payload of the message
    @type : int (-1 for a single byte message)

    @att payload : payload of the message
    @type : list of integers

    @att checksum : checked checksum
    @type : True or False

    """

    def __init__(self, mtype, lenpayload=None):
        """ create an instance of LCMessageReceived  on the fly (for received messages)

        @param mtype : mtype of the message
        @type : ascii character

        @param lenpayload : length of the payload of the message
        @type : integer or None"""
        self.datadict = {}
        self.received = True
        if mtype is None:
            LCMessageError(f'{mtype} must be present ')
            return
        else:
            self.mtype = mtype
        self._payload = Payload()
        self.lenpayload = None
        # index used to create the dictionnary from the payload
        self.payloadindex = 0
        if self.mtype.isupper():
            self.lenpayload = -1
            self._checksum = None
            return
        if self.mtype.islower():
            if isinstance(lenpayload, int):
                self.lenpayload = lenpayload
                return
            elif lenpayload is not None:
                LCMessageError(f'{lenpayload} must be None or integer', self)
                return

    def _decode2bytes(self, bytes2):
            """ decode 2 bytes representing a single value
            right most bit of byte1 bealongs to byte2"""
            if (isinstance(bytes2, list)
                and len(bytes2) == 2
                and isinstance(bytes2[0], int)
                and isinstance(bytes2[1], int)
                ):
                # least significant bit of byte1 is to put in bytes2
                t = bytes2[0] % 2
                lowbyte = bytes2[1] + t * 128
                highbyte = (bytes2[0] >> 1)
                res = highbyte * 256 + lowbyte
                return res
            else:
                LCMessageError(f'in decode2bytes : {bytes2} cannot be decoded', self)

    def _decodebits(self, byte1, bits):
        """ decode a byte containing several individual bits
        returns the list of individual values the first one
        being the value of the list significant bit"""
        if isinstance(byte1, int) and isinstance(bits, int):
            res = []
            for i in range(0, bits):
                res.append(byte1 % 2)
                byte1 = byte1 >> 1
            res.reverse()
            return res

    @staticmethod
    def _decodebyte(byte):
        """ return the byte, length is not used"""

        return byte

    @staticmethod
    def _decodeRTLBuff(ints):
        char = []

        pass

    @staticmethod
    def _decodeLTRBuff(ints):
        char = []
        for i in range(len(ints), 0, -1):
            num = ints[i - 1]
            for b in range(0, 7):
                char.extend(str(num % 2))
                num = num >> 1
        res = ''.join(char[::-1])
        return res

    def _getValue(self, vtype=None, length=None, extra=None, res1=None, res2=None):
        """builds the dictionnary entry for a value

        @param vtype : type of the value
        @type : string, amongst the keys of the mesvaluetypes dict in common.LCConstants

        @param length : for certain types the length of the value
        @type : integer

        @param : an extra function that is to be called after getting the value to get the final value
        @type : string"""

        if self.payloadindex < self.lenpayload:
            i = self.payloadindex
            if vtype == 'byte':
                value = self._decodebyte(self.payload[i])
                self.payloadindex += 1
            elif vtype == '2bytes':
                value = self._decode2bytes(self.payload[i: i + 2])
                self.payloadindex += 2
            elif vtype == 'strtype':
                value = self.payload[i: i + length]
                value = ''.join((chr(e) for e in value))
                self.payloadindex += length
            elif vtype == 'bits':
                value = self._decodebits(self.payload[i], length)
                self.payloadindex += 1
            elif vtype == 'bool':
                value = value
                self.payloadindex += 1
            elif vtype == 'LTRBuff':
                value = self._decodeLTRBuff(self.payload[i: i + length])
                self.payloadindex += length
            else:
                LCMessageError(f'{value} has no type definition', self)
            if extra != None:
                value = self._extra(extra, value)
                if value is None:
                    LCMessageError(f'invalid message value : '
                                   f'{value} does not comply with {extra} function {self.strucLogMessage()}', self)

            return value
        # the value is optional and not present
        else:
            return None

    def createLCmesDict(self):
        """ creates LCmessage.datadict  containing the messages values
        returns LCmessage.datadict"""
        # with mtype gets the dictionnary describing the values of the message
        coding = LCMesConstants.rMessagesDef.get(self.mtype, None)
        if coding == None or not self.isValid():
            # ignore invalid messages
            LCMessageError(f'invalid message payload mtype: {self.mtype}'
                           f'payload : {self.payload}', self)
            return
        res = {}
        res.update({'class': coding.get('class')})

        # values is the list of values
        values = coding.get('values')
        if values == None:
            # single byte message, done
            self.datadict = res
            return res

        for value in values:
            kwargs = LCMesConstants.mesTypesDef.get(value, None)
            if kwargs == None:
                LCMessageError(f'{value} has no type definition', self)
                return
            elif value == 'RTLBuffer' or value == 'LTRBuffer':
                left = res.get('leftmost')
                right = res.get('rightmost')
                positions = abs(left - right)
                extrabits = 7 - (abs(left - right) % 7) if (left - right) % 7 != 0 else 0
                kwargs['length'] = abs(int((left - right) / 7)) + 1 * ((left - right) % 7 != 0)
                sliceobjet = slice(extrabits, None, None)
                kwargs.update({'extra': {'sliceO': sliceobjet}})
            v = self._getValue(**kwargs)
            if v is None and self.mtype == 'e':
                # an optional message number is not present
                    errorcode = res.get('errorcode')
                    if errorcode in LCMesConstants.eCodesWithMesNum:
                        LCMessageError(f'error message {errorcode} should have message number', self)
            if isinstance(v, dict):
                res.update(v)
            elif isinstance(v, list) or isinstance(v, tuple):
                res.update({str(value): v[0]})
            else:
                res.update({str(value): v})
        self.datadict = res
        return res

    def createEvent(self):
        """ returns an event containing
        evname , LCmessage datadict , mtype and checksum"""

        self.createLCmesDict()
        name = 'on_msg_in_' + self.mtype
        if self.mtype == 'e':
            code = [LCMesConstants.eventCodes[index] for index, value
                    in enumerate(LCMesConstants.alleCodes) if value == self.datadict.get('errorcode')]
            name = 'on_msg_in_' + self.mtype + '_' + code[0]

        newdic = self.datadict
        newdic.update({'evname': name, 'mtype': self.mtype, 'checksum': self.checksum})
        return newdic


if __name__ == '__main__':

    import unittest
    import time
    from unittest.mock import patch, Mock
    from tests.LC_model.tests_LC_message_received import TestCreateLCMessageReceived, TestCreateLCmesDict

    unittest.main()


